# This file is part of ssh2-python.
# Copyright (C) 2020 Red_M

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import enum

cimport error_codes
cimport c_ssh2
cimport c_sftp


class Session(enum.Enum):
    BLOCK_INBOUND = c_ssh2.LIBSSH2_SESSION_BLOCK_INBOUND
    BLOCK_OUTBOUND = c_ssh2.LIBSSH2_SESSION_BLOCK_OUTBOUND


class Hostkey(enum.Enum):
    HASH_MD5 = c_ssh2.LIBSSH2_HOSTKEY_HASH_MD5
    HASH_SHA1 = c_ssh2.LIBSSH2_HOSTKEY_HASH_SHA1
    TYPE_UNKNOWN = c_ssh2.LIBSSH2_HOSTKEY_TYPE_UNKNOWN
    TYPE_RSA = c_ssh2.LIBSSH2_HOSTKEY_TYPE_RSA
    TYPE_DSS = c_ssh2.LIBSSH2_HOSTKEY_TYPE_DSS
    TYPE_ECDSA_256 = c_ssh2.LIBSSH2_HOSTKEY_TYPE_ECDSA_256
    TYPE_ECDSA_384 = c_ssh2.LIBSSH2_HOSTKEY_TYPE_ECDSA_384
    TYPE_ECDSA_521 = c_ssh2.LIBSSH2_HOSTKEY_TYPE_ECDSA_521
    TYPE_ED25519 = c_ssh2.LIBSSH2_HOSTKEY_TYPE_ED25519


class Callback(enum.Enum):
    RECV = c_ssh2.LIBSSH2_CALLBACK_RECV
    SEND = c_ssh2.LIBSSH2_CALLBACK_SEND
    X11 = c_ssh2.LIBSSH2_CALLBACK_X11


class Method(enum.Enum):
    KEX = c_ssh2.LIBSSH2_METHOD_KEX
    HOSTKEY = c_ssh2.LIBSSH2_METHOD_HOSTKEY
    CRYPT_CS = c_ssh2.LIBSSH2_METHOD_CRYPT_CS
    CRYPT_SC = c_ssh2.LIBSSH2_METHOD_CRYPT_SC
    MAC_CS = c_ssh2.LIBSSH2_METHOD_MAC_CS
    MAC_SC = c_ssh2.LIBSSH2_METHOD_MAC_SC
    COMP_CS = c_ssh2.LIBSSH2_METHOD_COMP_CS
    COMP_SC = c_ssh2.LIBSSH2_METHOD_COMP_SC
    LANG_CS = c_ssh2.LIBSSH2_METHOD_LANG_CS
    LANG_SC = c_ssh2.LIBSSH2_METHOD_LANG_SC


class Flag(enum.Enum):
    SIGPIPE = c_ssh2.LIBSSH2_FLAG_SIGPIPE
    COMPRESS = c_ssh2.LIBSSH2_FLAG_COMPRESS


class SFTP(enum.Enum):
    # Type of file mask
    S_IFMT = c_sftp.LIBSSH2_SFTP_S_IFMT
    # named pipe (fifo)
    S_IFIFO = c_sftp.LIBSSH2_SFTP_S_IFIFO
    # character special
    S_IFCHR = c_sftp.LIBSSH2_SFTP_S_IFCHR
    # directory
    S_IFDIR = c_sftp.LIBSSH2_SFTP_S_IFDIR
    # block special (block device)
    S_IFBLK = c_sftp.LIBSSH2_SFTP_S_IFBLK
    # regular
    S_IFREG = c_sftp.LIBSSH2_SFTP_S_IFREG
    # symbolic link
    S_IFLNK = c_sftp.LIBSSH2_SFTP_S_IFLNK
    # socket
    S_IFSOCK = c_sftp.LIBSSH2_SFTP_S_IFSOCK

    # File Transfer Flags
    FXF_READ = c_sftp.LIBSSH2_FXF_READ
    FXF_WRITE = c_sftp.LIBSSH2_FXF_WRITE
    FXF_APPEND = c_sftp.LIBSSH2_FXF_APPEND
    FXF_CREAT = c_sftp.LIBSSH2_FXF_CREAT
    FXF_TRUNC = c_sftp.LIBSSH2_FXF_TRUNC
    FXF_EXCL = c_sftp.LIBSSH2_FXF_EXCL

    # File mode masks
    # Read, write, execute/search by owner
    S_IRWXU = c_sftp.LIBSSH2_SFTP_S_IRWXU
    S_IRUSR = c_sftp.LIBSSH2_SFTP_S_IRUSR
    S_IWUSR = c_sftp.LIBSSH2_SFTP_S_IWUSR
    S_IXUSR = c_sftp.LIBSSH2_SFTP_S_IXUSR
    # Read, write, execute/search by group
    S_IRWXG = c_sftp.LIBSSH2_SFTP_S_IRWXG
    S_IRGRP = c_sftp.LIBSSH2_SFTP_S_IRGRP
    S_IWGRP = c_sftp.LIBSSH2_SFTP_S_IWGRP
    S_IXGRP = c_sftp.LIBSSH2_SFTP_S_IXGRP
    # Read, write, execute/search by others
    S_IRWXO = c_sftp.LIBSSH2_SFTP_S_IRWXO
    S_IROTH = c_sftp.LIBSSH2_SFTP_S_IROTH
    S_IWOTH = c_sftp.LIBSSH2_SFTP_S_IWOTH
    S_IXOTH = c_sftp.LIBSSH2_SFTP_S_IXOTH

    # Read only
    ST_RDONLY = c_sftp.LIBSSH2_SFTP_ST_RDONLY
    # No suid
    ST_NOSUID = c_sftp.LIBSSH2_SFTP_ST_NOSUID


class KnownHost(enum.Enum):
    # Host format type masks
    TYPE_MASK = c_ssh2.LIBSSH2_KNOWNHOST_TYPE_MASK
    TYPE_PLAIN = c_ssh2.LIBSSH2_KNOWNHOST_TYPE_PLAIN
    TYPE_SHA1 = c_ssh2.LIBSSH2_KNOWNHOST_TYPE_SHA1
    TYPE_CUSTOM = c_ssh2.LIBSSH2_KNOWNHOST_TYPE_CUSTOM

    # Key format type masks
    KEYENC_MASK = c_ssh2.LIBSSH2_KNOWNHOST_KEYENC_MASK
    KEYENC_RAW = c_ssh2.LIBSSH2_KNOWNHOST_KEYENC_RAW
    KEYENC_BASE64 = c_ssh2.LIBSSH2_KNOWNHOST_KEYENC_BASE64

    # Key type masks
    KEY_MASK = c_ssh2.LIBSSH2_KNOWNHOST_KEY_MASK
    KEY_SHIFT = c_ssh2.LIBSSH2_KNOWNHOST_KEY_SHIFT
    KEY_RSA1 = c_ssh2.LIBSSH2_KNOWNHOST_KEY_RSA1
    KEY_SSHRSA = c_ssh2.LIBSSH2_KNOWNHOST_KEY_SSHRSA
    KEY_SSHDSS = c_ssh2.LIBSSH2_KNOWNHOST_KEY_SSHDSS
    IF EMBEDDED_LIB:
        KEY_UNKNOWN = c_ssh2.LIBSSH2_KNOWNHOST_KEY_UNKNOWN


class ErrorCodes(enum.Enum):
    NONE = error_codes._LIBSSH2_ERROR_NONE
    SOCKET_NONE = error_codes._LIBSSH2_ERROR_SOCKET_NONE
    BANNER_RECV = error_codes._LIBSSH2_ERROR_BANNER_RECV
    BANNER_SEND = error_codes._LIBSSH2_ERROR_BANNER_SEND
    KEY_EXCHANGE_FAILURE \
        = error_codes._LIBSSH2_ERROR_KEY_EXCHANGE_FAILURE
    TIMEOUT = error_codes._LIBSSH2_ERROR_TIMEOUT
    HOSTKEY_INIT = error_codes._LIBSSH2_ERROR_HOSTKEY_INIT
    HOSTKEY_SIGN = error_codes._LIBSSH2_ERROR_HOSTKEY_SIGN
    DECRYPT = error_codes._LIBSSH2_ERROR_DECRYPT
    SOCKET_DISCONNECT = error_codes._LIBSSH2_ERROR_SOCKET_DISCONNECT
    PROTO = error_codes._LIBSSH2_ERROR_PROTO
    PASSWORD_EXPIRED = error_codes._LIBSSH2_ERROR_PASSWORD_EXPIRED
    FILE = error_codes._LIBSSH2_ERROR_FILE
    METHOD_NONE \
        = error_codes._LIBSSH2_ERROR_METHOD_NONE
    AUTHENTICATION_FAILED \
        = error_codes._LIBSSH2_ERROR_AUTHENTICATION_FAILED
    PUBLICKEY_UNRECOGNIZED \
        = error_codes._LIBSSH2_ERROR_PUBLICKEY_UNRECOGNIZED
    PUBLICKEY_UNVERIFIED \
        = error_codes._LIBSSH2_ERROR_PUBLICKEY_UNVERIFIED
    CHANNEL_OUTOFORDER = error_codes._LIBSSH2_ERROR_CHANNEL_OUTOFORDER
    CHANNEL_FAILURE = error_codes._LIBSSH2_ERROR_CHANNEL_FAILURE
    CHANNEL_REQUEST_DENIED \
        = error_codes._LIBSSH2_ERROR_CHANNEL_REQUEST_DENIED
    CHANNEL_UNKNOWN = error_codes._LIBSSH2_ERROR_CHANNEL_UNKNOWN
    CHANNEL_WINDOW_EXCEEDED \
        = error_codes._LIBSSH2_ERROR_CHANNEL_WINDOW_EXCEEDED
    CHANNEL_PACKET_EXCEEDED \
        = error_codes._LIBSSH2_ERROR_CHANNEL_PACKET_EXCEEDED
    CHANNEL_CLOSED = error_codes._LIBSSH2_ERROR_CHANNEL_CLOSED
    CHANNEL_EOF_SENT = error_codes._LIBSSH2_ERROR_CHANNEL_EOF_SENT
    SCP_PROTOCOL = error_codes._LIBSSH2_ERROR_SCP_PROTOCOL
    ZLIB = error_codes._LIBSSH2_ERROR_ZLIB
    SOCKET_TIMEOUT = error_codes._LIBSSH2_ERROR_SOCKET_TIMEOUT
    SFTP_PROTOCOL = error_codes._LIBSSH2_ERROR_SFTP_PROTOCOL
    REQUEST_DENIED = error_codes._LIBSSH2_ERROR_REQUEST_DENIED
    METHOD_NOT_SUPPORTED \
        = error_codes._LIBSSH2_ERROR_METHOD_NOT_SUPPORTED
    INVAL = error_codes._LIBSSH2_ERROR_INVAL
    INVALID_POLL_TYPE = error_codes._LIBSSH2_ERROR_INVALID_POLL_TYPE
    PUBLICKEY_PROTOCOL = error_codes._LIBSSH2_ERROR_PUBLICKEY_PROTOCOL
    EAGAIN = error_codes._LIBSSH2_ERROR_EAGAIN
    LIBSSH2CHANNEL_EAGAIN = error_codes._LIBSSH2CHANNEL_EAGAIN
    BUFFER_TOO_SMALL = error_codes._LIBSSH2_ERROR_BUFFER_TOO_SMALL
    BAD_USE = error_codes._LIBSSH2_ERROR_BAD_USE
    COMPRESS = error_codes._LIBSSH2_ERROR_COMPRESS
    OUT_OF_BOUNDARY = error_codes._LIBSSH2_ERROR_OUT_OF_BOUNDARY
    AGENT_PROTOCOL = error_codes._LIBSSH2_ERROR_AGENT_PROTOCOL
    SOCKET_RECV = error_codes._LIBSSH2_ERROR_SOCKET_RECV
    SOCKET_SEND = error_codes._LIBSSH2_ERROR_SOCKET_SEND
    ENCRYPT = error_codes._LIBSSH2_ERROR_ENCRYPT
    BAD_SOCKET = error_codes._LIBSSH2_ERROR_BAD_SOCKET
    KEX_FAILURE = error_codes._LIBSSH2_ERROR_KEX_FAILURE
    INVALID_MAC = error_codes._LIBSSH2_ERROR_INVALID_MAC
    IF EMBEDDED_LIB:
        KNOWN_HOSTS = error_codes._LIBSSH2_ERROR_KNOWN_HOSTS



# This file is part of ssh2-python.
# Copyright (C) 2017 Panos Kittenis

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""Available only when built on libssh2 >= ``1.7``"""

from libc.stdlib cimport malloc, free

cimport c_ssh2


IF EMBEDDED_LIB:
    cdef class FileInfo:
        """Representation of stat structure - libssh2 >= 1.7"""

        def __cinit__(self):
            with nogil:
                self._stat = <c_ssh2.libssh2_struct_stat *>malloc(
                    sizeof(c_ssh2.libssh2_struct_stat))
                if self._stat is NULL:
                    with gil:
                        raise MemoryError

        def __dealloc__(self):
            free(self._stat)

        @property
        def st_size(self):
            return self._stat.st_size

        @property
        def st_mode(self):
            return self._stat.st_mode

        IF UNAME_SYSNAME != "Windows":
            @property
            def st_ino(self):
                return self._stat.st_ino

            @property
            def st_nlink(self):
                return self._stat.st_nlink

            @property
            def st_uid(self):
                return self._stat.st_uid

            @property
            def st_gid(self):
                return self._stat.st_gid

            @property
            def st_rdev(self):
                return self._stat.st_rdev

            @property
            def st_blksize(self):
                return self._stat.st_blksize

            @property
            def st_blocks(self):
                return self._stat.st_blocks

            @property
            def st_atime(self):
                return self._stat.st_atime

            @property
            def st_mtime(self):
                return self._stat.st_mtime

            @property
            def st_ctime(self):
                return self._stat.st_ctime



import os

from sys import stderr
from subprocess import check_call
from glob import glob
from shutil import copy2
from multiprocessing import cpu_count


def build_ssh2():
    SYSTEM_BUILD_MINGW = bool(os.environ.get('SYSTEM_BUILD_MINGW', 0))
    if bool(os.environ.get('SYSTEM_LIBSSH', False)):
        stderr.write("Using system libssh2..%s" % (os.sep))
        return
    if os.path.exists('/usr/local/opt/openssl'):
        os.environ['OPENSSL_ROOT_DIR'] = '/usr/local/opt/openssl'

    if not os.path.exists('src'):
        os.mkdir('src')

    os.chdir('src')
    if SYSTEM_BUILD_MINGW==1:
        check_call('cmake -DCMAKE_TOOLCHAIN_FILE=../ci/gitlab/windows/windows_toolchain.cmake \
        ../libssh2 -DBUILD_SHARED_LIBS=ON \
        -DENABLE_ZLIB_COMPRESSION=ON -DENABLE_CRYPT_NONE=ON \
        -DENABLE_MAC_NONE=ON -DCRYPTO_BACKEND=OpenSSL',shell=True, env=os.environ)
    else:
        check_call('cmake ../libssh2 -DBUILD_SHARED_LIBS=ON \
        -DENABLE_ZLIB_COMPRESSION=ON -DENABLE_CRYPT_NONE=ON \
        -DENABLE_MAC_NONE=ON -DCRYPTO_BACKEND=OpenSSL',shell=True, env=os.environ)
    check_call('cmake --build . --config Release', shell=True, env=os.environ)
    os.chdir('..')

    if SYSTEM_BUILD_MINGW==1:
        for src in glob('src/src/libssh2.dll'):
            copy2(src, 'ssh2/')
    else:
        for src in glob('src/src/libssh2.so*'):
            copy2(src, 'ssh2/')



from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from . import session
from . import channel
from . import sftp
from . import exceptions
from . import knownhost
from . import agent
from . import listener
from . import pkey as privatekey
from . import publickey
from . import sftp_handle
from . import fileinfo
from . import statinfo
from . import utils
from . import libssh2_enums as enums



# This file is part of ssh2-python.
# Copyright (C) 2017 Panos Kittenis

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import error_codes
from cpython cimport PyObject_AsFileDescriptor
from libc.stdlib cimport malloc, free
from libc.time cimport time_t
from cython.operator cimport dereference as c_dereference
from libc.string cimport strlen, strdup


from agent cimport PyAgent, agent_auth, agent_init, init_connect_agent
from channel cimport PyChannel
from exceptions import SessionHostKeyError, KnownHostError, \
    PublicKeyInitError, ChannelError
from listener cimport PyListener
from sftp cimport PySFTP
from publickey cimport PyPublicKeySystem
from utils cimport to_bytes, to_str, handle_error_codes
from statinfo cimport StatInfo
from knownhost cimport PyKnownHost
from fileinfo cimport FileInfo


cimport c_ssh2
cimport c_sftp
cimport c_pkey


LIBSSH2_SESSION_BLOCK_INBOUND = c_ssh2.LIBSSH2_SESSION_BLOCK_INBOUND
LIBSSH2_SESSION_BLOCK_OUTBOUND = c_ssh2.LIBSSH2_SESSION_BLOCK_OUTBOUND
LIBSSH2_HOSTKEY_HASH_MD5 = c_ssh2.LIBSSH2_HOSTKEY_HASH_MD5
LIBSSH2_HOSTKEY_HASH_SHA1 = c_ssh2.LIBSSH2_HOSTKEY_HASH_SHA1
LIBSSH2_HOSTKEY_TYPE_UNKNOWN = c_ssh2.LIBSSH2_HOSTKEY_TYPE_UNKNOWN
LIBSSH2_HOSTKEY_TYPE_RSA = c_ssh2.LIBSSH2_HOSTKEY_TYPE_RSA
LIBSSH2_HOSTKEY_TYPE_DSS = c_ssh2.LIBSSH2_HOSTKEY_TYPE_DSS

LIBSSH2_CALLBACK_RECV = c_ssh2.LIBSSH2_CALLBACK_RECV
LIBSSH2_CALLBACK_SEND = c_ssh2.LIBSSH2_CALLBACK_SEND
LIBSSH2_CALLBACK_X11 = c_ssh2.LIBSSH2_CALLBACK_X11

LIBSSH2_METHOD_KEX = c_ssh2.LIBSSH2_METHOD_KEX
LIBSSH2_METHOD_HOSTKEY = c_ssh2.LIBSSH2_METHOD_HOSTKEY
LIBSSH2_METHOD_CRYPT_CS = c_ssh2.LIBSSH2_METHOD_CRYPT_CS
LIBSSH2_METHOD_CRYPT_SC = c_ssh2.LIBSSH2_METHOD_CRYPT_SC
LIBSSH2_METHOD_MAC_CS = c_ssh2.LIBSSH2_METHOD_MAC_CS
LIBSSH2_METHOD_MAC_SC = c_ssh2.LIBSSH2_METHOD_MAC_SC
LIBSSH2_METHOD_COMP_CS = c_ssh2.LIBSSH2_METHOD_COMP_CS
LIBSSH2_METHOD_COMP_SC = c_ssh2.LIBSSH2_METHOD_COMP_SC
LIBSSH2_METHOD_LANG_CS = c_ssh2.LIBSSH2_METHOD_LANG_CS
LIBSSH2_METHOD_LANG_SC = c_ssh2.LIBSSH2_METHOD_LANG_SC

LIBSSH2_FLAG_SIGPIPE = c_ssh2.LIBSSH2_FLAG_SIGPIPE
LIBSSH2_FLAG_COMPRESS = c_ssh2.LIBSSH2_FLAG_COMPRESS


cdef void kbd_callback(const char *name, int name_len,
                       const char *instruction, int instruction_len,
                       int num_prompts,
                       const c_ssh2.LIBSSH2_USERAUTH_KBDINT_PROMPT *prompts,
                       c_ssh2.LIBSSH2_USERAUTH_KBDINT_RESPONSE *responses,
                       void **abstract) except *:
    py_sess = (<Session>c_dereference(abstract))
    if py_sess._kbd_callback is None:
        return
    cdef bytes b_password = to_bytes(py_sess._kbd_callback())
    cdef char *_password = b_password
    if num_prompts==1:
        responses[0].text = strdup(_password)
        responses[0].length = strlen(_password)


cdef class Session:

    """LibSSH2 Session class providing session functions"""

    def __cinit__(self):
        self._session = c_ssh2.libssh2_session_init_ex(NULL,
                                                       NULL, NULL, <void*> self)
        if self._session is NULL:
            raise MemoryError
        self._sock = 0
        self.sock = None
        self._kbd_callback = None

    def __dealloc__(self):
        if self._session is not NULL:
            c_ssh2.libssh2_session_free(self._session)
        self._session = NULL

    def disconnect(self):
        cdef int rc
        with nogil:
            rc = c_ssh2.libssh2_session_disconnect(self._session, b"end")
        return handle_error_codes(rc)

    def method_pref(self, method_type, pref_methods):
        """Set internal perferences based on ``method_type`` to
        ``pref_methods``.

        Valid ``method_type`` options are:

        - ``LIBSSH2_METHOD_KEX``
            For key exchange.

        - ``LIBSSH2_METHOD_HOSTKEY``
            For selecting host key type.

        - ``LIBSSH2_METHOD_CRYPT_CS``
            Encryption between client to server

        - ``LIBSSH2_METHOD_CRYPT_SC``
            Encryption between server to client

        - ``LIBSSH2_METHOD_MAC_CS``
            MAC between client to server

        - ``LIBSSH2_METHOD_MAC_SC``
            MAC between server to client

        - ``LIBSSH2_METHOD_COMP_CS``
            Compression between client to server

        - ``LIBSSH2_METHOD_COMP_SC``
            Compression between server to client

        - ``LIBSSH2_METHOD_LANG_CS``
            Language between client to server

        - ``LIBSSH2_METHOD_LANG_SC``
            Language between server to client


        Valid options that end in ``CS`` are from the client to the server and
        the inverse is true as well.

        Valid ``pref_methods`` options are dependant on the ``method_type``
        selected. Refer to the libssh2 docs

        Must be called before ``self.handshake()`` if you wish to change
        the defaults.

        Return 0 on success or negative on failure.
        It returns ``ssh2.error_codes.LIBSSH2_ERROR_EAGAIN`` when it would
        otherwise block.
        While ``ssh2.error_codes.LIBSSH2_ERROR_EAGAIN`` is a negative number,
        it isn't
        really a failure per se.

        :raises: :py:class:`ssh2.exceptions.MethodNotSupported` on an incorrect
          ``method_type`` or ``pref_methods`` argument(s).
        :param method_type: Method perference to change.
        :type method_type: ``ssh2.session.LIBSSH2_METHOD_*``
        :param pref_methods: Coma delimited list as a bytes string of preferred
        methods to use with the most preferred listed first and the least
        preferred listed last.
        If a method is listed which is not supported by libssh2 it will be
        ignored and not sent to the remote host during protocol negotiation.
        :type pref_methods: bytes
        :rtype: ``int``
        """
        if (int(method_type)<1 and int(method_type)>10) or self._sock!=0:
            return handle_error_codes(
                error_codes.LIBSSH2_ERROR_METHOD_NOT_SUPPORTED)
        cdef int rc
        cdef int _method_type = int(method_type)
        cdef bytes b_pref_methods = to_bytes(pref_methods)
        cdef char *_pref_methods = b_pref_methods
        with nogil:
            rc = c_ssh2.libssh2_session_method_pref(self._session,
                                                    _method_type, _pref_methods)
        return handle_error_codes(rc)

    def methods(self, method_type):
        """Get internal perferences used to negotiate based on ``method_type``.

        Valid ``method_type`` options are:

        - ``LIBSSH2_METHOD_KEX``
            For key exchange.

        - ``LIBSSH2_METHOD_HOSTKEY``
            For selecting host key type.

        - ``LIBSSH2_METHOD_CRYPT_CS``
            Encryption between client to server

        - ``LIBSSH2_METHOD_CRYPT_SC``
            Encryption between server to client

        - ``LIBSSH2_METHOD_MAC_CS``
            MAC between client to server

        - ``LIBSSH2_METHOD_MAC_SC``
            MAC between server to client

        - ``LIBSSH2_METHOD_COMP_CS``
            Compression between client to server

        - ``LIBSSH2_METHOD_COMP_SC``
            Compression between server to client

        - ``LIBSSH2_METHOD_LANG_CS``
            Language between client to server

        - ``LIBSSH2_METHOD_LANG_SC``
            Language between server to client


        Valid options that end in ``CS`` are from the client to the server and
        the inverse is true as well.

        :raises: :py:class:`ssh2.exceptions.MethodNotSupported` on an incorrect
          ``method_type`` argument.
        :param method_type: Method type.
        :type method_type: ``ssh2.session.LIBSSH2_METHOD_*``
        :rtype: ``bytes``
        """
        if (int(method_type)<1 and int(method_type)>10) or self._sock==0:
            return handle_error_codes(
                error_codes.LIBSSH2_ERROR_METHOD_NOT_SUPPORTED)
        cdef const char *ret
        cdef int _method_type = int(method_type)
        with nogil:
            ret = c_ssh2.libssh2_session_methods(self._session, _method_type)
        return ret

    def supported_algs(self, method_type, algs):
        """Get the supported internal perferences based on ``method_type`` and
        ``algs``.

        Valid ``method_type`` options are:

        - ``LIBSSH2_METHOD_KEX``
            For key exchange.

        - ``LIBSSH2_METHOD_HOSTKEY``
            For selecting host key type.

        - ``LIBSSH2_METHOD_CRYPT_CS``
            Encryption between client to server

        - ``LIBSSH2_METHOD_CRYPT_SC``
            Encryption between server to client

        - ``LIBSSH2_METHOD_MAC_CS``
            MAC between client to server

        - ``LIBSSH2_METHOD_MAC_SC``
            MAC between server to client

        - ``LIBSSH2_METHOD_COMP_CS``
            Compression between client to server

        - ``LIBSSH2_METHOD_COMP_SC``
            Compression between server to client

        - ``LIBSSH2_METHOD_LANG_CS``
            Language between client to server

        - ``LIBSSH2_METHOD_LANG_SC``
            Language between server to client


        :raises: :py:class:`ssh2.exceptions.MethodNotSupported` on an incorrect
          ``method_type`` or ``algs`` argument(s).
        :param method_type: Method type.
        :type method_type: ``ssh2.session.LIBSSH2_METHOD_*``
        :param algs: Coma delimited list as a bytes string.
        :type algs: ``bytes str``
        :rtype: ``array``
        """
        if (int(method_type)<1 and int(method_type)>10) or \
                isinstance(algs, type(b'')) is False:
            return handle_error_codes(
                error_codes.LIBSSH2_ERROR_METHOD_NOT_SUPPORTED)
        out = []
        # algs = to_bytes(algs)
        algs = algs.split(b',')
        cdef int i, j = 0
        cdef int rc
        cdef int _method_type = int(method_type)
        cdef const char **_algs = \
            <const char**> malloc(len(algs) * sizeof(char*))
        for item in algs:
            _algs[j] = item
            j+=1

        with nogil:
            rc = c_ssh2.libssh2_session_supported_algs(self._session,
                                                       _method_type, &_algs)

        if rc>0:
            while i<rc:
                out.append(algs[i])
                i+=1
            with nogil:
                c_ssh2.libssh2_free(self._session, _algs)
            return out
        else:
            return handle_error_codes(rc)

    def flag(self, set_flag, value):
        """Set options for the session.

        ``set_flag`` is the option to set, while ``value`` is typically set to
        ``1`` or ``0`` to enable or disable the option.

        Valid flags are:

        - ``ssh2.session.LIBSSH2_FLAG_SIGPIPE``
            If set, libssh2 will not attempt to block SIGPIPEs but will let them
            trigger from the underlying socket layer.
        - ``ssh2.session.LIBSSH2_FLAG_COMPRESS``
            If set - before the connection negotiation is performed -
            libssh2 will try to negotiate compression enabling for this
            connection. By default libssh2 will not attempt to use compression.

        Must be called before ``self.handshake()`` if you wish to change
        options.


        :raises: :py:class:`ssh2.exceptions.MethodNotSupported` on an incorrect
          ``flag`` or ``value`` argument(s).
        :param set_flag: Flag to set. See above for options.
        :type set_flag: ``ssh2.session.LIBSSH2_METHOD_*``
        :param value: Value that ``set_flag`` will be set to. Must be ``0`` or
        ``1``.
        :type value: ``int``
        :rtype: ``int``
        """
        if (int(set_flag)<1 or int(set_flag)>3) or (value!=0 and value!=1):
            return handle_error_codes(
                error_codes.LIBSSH2_ERROR_METHOD_NOT_SUPPORTED)
        cdef int rc
        cdef int _set_flag = int(set_flag)
        cdef int _value = int(value)
        with nogil:
            rc = c_ssh2.libssh2_session_flag(self._session, _set_flag, _value)
        return handle_error_codes(rc)

    def handshake(self, sock not None):
        """Perform SSH handshake.

        Must be called after Session initialisation."""
        cdef int _sock = PyObject_AsFileDescriptor(sock)
        cdef int rc
        with nogil:
            rc = c_ssh2.libssh2_session_handshake(self._session, _sock)
            self._sock = _sock
        self.sock = sock
        return handle_error_codes(rc)

    def startup(self, sock):
        """Deprecated - use self.handshake"""
        cdef int _sock = PyObject_AsFileDescriptor(sock)
        cdef int rc
        rc = c_ssh2.libssh2_session_startup(self._session, _sock)
        return handle_error_codes(rc)

    def set_blocking(self, bint blocking):
        """Set session blocking mode on/off.

        :param blocking: ``False`` for non-blocking, ``True`` for blocking.
          Session default is blocking unless set otherwise.
        :type blocking: bool"""
        with nogil:
            c_ssh2.libssh2_session_set_blocking(
                self._session, blocking)

    def get_blocking(self):
        """Get session blocking mode enabled True/False.

        :rtype: bool"""
        cdef int rc
        with nogil:
            rc = c_ssh2.libssh2_session_get_blocking(self._session)
        return bool(rc)

    def set_timeout(self, long timeout):
        """Set the timeout in milliseconds for how long a blocking
        call may wait until the situation is considered an error and
        :py:class:`ssh2.error_codes.LIBSSH2_ERROR_TIMEOUT` is returned.

        By default or if timeout set is zero, blocking calls do not
        time out.
        :param timeout: Milliseconds to wait before timeout."""
        with nogil:
            c_ssh2.libssh2_session_set_timeout(self._session, timeout)

    def get_timeout(self):
        """Get current session timeout setting"""
        cdef long timeout
        with nogil:
            timeout = c_ssh2.libssh2_session_get_timeout(self._session)
        return timeout

    def userauth_authenticated(self):
        """True/False for is user authenticated or not.

        :rtype: bool"""
        cdef bint rc
        with nogil:
            rc = c_ssh2.libssh2_userauth_authenticated(self._session)
        return bool(rc)

    def userauth_list(self, username not None):
        """Retrieve available authentication methods list.

        :rtype: list"""
        cdef bytes b_username = to_bytes(username)
        cdef char *_username = b_username
        cdef size_t username_len = len(b_username)
        cdef char *_auth
        cdef str auth
        with nogil:
            _auth = c_ssh2.libssh2_userauth_list(
                self._session, _username, username_len)
        if _auth is NULL:
            return
        auth = to_str(_auth)
        return auth.split(',')

    def userauth_publickey_fromfile(self, username not None,
                                    privatekey not None,
                                    passphrase='',
                                    publickey=None):
        """Authenticate with public key from file.

        :rtype: int"""
        cdef int rc
        cdef bytes b_username = to_bytes(username)
        cdef bytes b_publickey = to_bytes(publickey) \
            if publickey is not None else None
        cdef bytes b_privatekey = to_bytes(privatekey)
        cdef bytes b_passphrase = to_bytes(passphrase)
        cdef char *_username = b_username
        cdef char *_publickey = NULL
        cdef char *_privatekey = b_privatekey
        cdef char *_passphrase = b_passphrase
        if b_publickey is not None:
            _publickey = b_publickey
        with nogil:
            rc = c_ssh2.libssh2_userauth_publickey_fromfile(
                self._session, _username, _publickey, _privatekey, _passphrase)
        return handle_error_codes(rc)

    def userauth_publickey(self, username not None,
                           bytes pubkeydata not None):
        """Perform public key authentication with provided public key data

        :param username: User name to authenticate as
        :type username: str
        :param pubkeydata: Public key data
        :type pubkeydata: bytes

        :rtype: int"""
        cdef int rc
        cdef bytes b_username = to_bytes(username)
        cdef char *_username = b_username
        cdef unsigned char *_pubkeydata = pubkeydata
        cdef size_t pubkeydata_len = len(pubkeydata)
        with nogil:
            rc = c_ssh2.libssh2_userauth_publickey(
                self._session, _username, _pubkeydata,
                pubkeydata_len, NULL, NULL)
        return handle_error_codes(rc)

    def userauth_hostbased_fromfile(self,
                                    username not None,
                                    privatekey not None,
                                    hostname not None,
                                    publickey=None,
                                    passphrase=''):
        cdef int rc
        cdef bytes b_username = to_bytes(username)
        cdef bytes b_publickey = to_bytes(publickey) \
            if publickey is not None else None
        cdef bytes b_privatekey = to_bytes(privatekey)
        cdef bytes b_passphrase = to_bytes(passphrase)
        cdef bytes b_hostname = to_bytes(hostname)
        cdef char *_username = b_username
        cdef char *_publickey = NULL
        cdef char *_privatekey = b_privatekey
        cdef char *_passphrase = b_passphrase
        cdef char *_hostname = b_hostname
        if b_publickey is not None:
            _publickey = b_publickey
        with nogil:
            rc = c_ssh2.libssh2_userauth_hostbased_fromfile(
                self._session, _username, _publickey,
                _privatekey, _passphrase, _hostname)
        return handle_error_codes(rc)

    IF EMBEDDED_LIB:
        def userauth_publickey_frommemory(
                self, username, bytes privatekeyfiledata,
                passphrase='', bytes publickeyfiledata=None):
            cdef int rc
            cdef bytes b_username = to_bytes(username)
            cdef bytes b_passphrase = to_bytes(passphrase)
            cdef char *_username = b_username
            cdef char *_passphrase = b_passphrase
            cdef char *_publickeyfiledata = NULL
            cdef char *_privatekeyfiledata = privatekeyfiledata
            cdef size_t username_len, pubkeydata_len, privatekeydata_len
            username_len, pubkeydata_len, privatekeydata_len = \
                len(b_username), 0, \
                len(privatekeyfiledata)
            if publickeyfiledata is not None:
                _publickeyfiledata = publickeyfiledata
                pubkeydata_len = len(publickeyfiledata)
            with nogil:
                rc = c_ssh2.libssh2_userauth_publickey_frommemory(
                    self._session, _username, username_len, _publickeyfiledata,
                    pubkeydata_len, _privatekeyfiledata,
                    privatekeydata_len, _passphrase)
            return handle_error_codes(rc)

    def userauth_password(self, username not None, password not None):
        """Perform password authentication

        :param username: User name to authenticate.
        :type username: str
        :param password: Password
        :type password: str"""
        cdef int rc
        cdef bytes b_username = to_bytes(username)
        cdef bytes b_password = to_bytes(password)
        cdef const char *_username = b_username
        cdef const char *_password = b_password
        with nogil:
            rc = c_ssh2.libssh2_userauth_password(
                self._session, _username, _password)
        return handle_error_codes(rc)

    def userauth_keyboardinteractive(self, username not None,
                                     password not None):
        """Perform keyboard-interactive authentication

        :param username: User name to authenticate.
        :type username: str
        :param password: Password
        :type password: str
        """
        cdef int rc
        cdef bytes b_username = to_bytes(username)
        cdef const char *_username = b_username

        def passwd():
            return(password)

        self._kbd_callback = passwd
        rc = c_ssh2.libssh2_userauth_keyboard_interactive(
            self._session, _username, &kbd_callback)
        self._kbd_callback = None
        return handle_error_codes(rc)

    def agent_init(self):
        """Initialise SSH agent.

        :rtype: :py:class:`ssh2.agent.Agent`
        """
        cdef c_ssh2.LIBSSH2_AGENT *agent
        with nogil:
            agent = agent_init(self._session)
        return PyAgent(agent, self)

    def agent_auth(self, username not None):
        """Convenience function for performing user authentication via SSH Agent.

        Initialises, connects to, gets list of identities from and attempts
        authentication with each identity from SSH agent.

        Note that agent connections cannot be used in non-blocking mode -
        clients should call `set_blocking(0)` *after* calling this function.

        On completion, or any errors, agent is disconnected and resources freed.

        All steps are performed in C space which makes this function perform
        better than calling the individual Agent class functions from
        Python.

        :raises: :py:class:`MemoryError` on error initialising agent
        :raises: :py:class:`ssh2.exceptions.AgentConnectionError` on error
          connecting to agent
        :raises: :py:class:`ssh2.exceptions.AgentListIdentitiesError` on error
          getting identities from agent
        :raises: :py:class:`ssh2.exceptions.AgentAuthenticationError` on no
          successful authentication with all available identities.
        :raises: :py:class:`ssh2.exceptions.AgentGetIdentityError` on error
          getting known identity from agent

        :rtype: None"""
        cdef bytes b_username = to_bytes(username)
        cdef char *_username = b_username
        cdef c_ssh2.LIBSSH2_AGENT *agent = NULL
        cdef c_ssh2.libssh2_agent_publickey *identity = NULL
        cdef c_ssh2.libssh2_agent_publickey *prev = NULL
        agent = init_connect_agent(self._session)
        with nogil:
            agent_auth(_username, agent)

    def open_session(self):
        """Open new channel session.

        :rtype: :py:class:`ssh2.channel.Channel`
        """
        cdef c_ssh2.LIBSSH2_CHANNEL *channel
        with nogil:
            channel = c_ssh2.libssh2_channel_open_session(
                self._session)
        if channel is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PyChannel(channel, self)

    def direct_tcpip_ex(self, host not None, int port,
                        shost not None, int sport):
        cdef c_ssh2.LIBSSH2_CHANNEL *channel
        cdef bytes b_host = to_bytes(host)
        cdef bytes b_shost = to_bytes(shost)
        cdef char *_host = b_host
        cdef char *_shost = b_shost
        with nogil:
            channel = c_ssh2.libssh2_channel_direct_tcpip_ex(
                self._session, _host, port, _shost, sport)
        if channel is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PyChannel(channel, self)

    def direct_tcpip(self, host not None, int port):
        """Open direct TCP/IP channel to host:port

        Channel will be listening on an available open port on client side
        as assigned by OS.
        """
        cdef c_ssh2.LIBSSH2_CHANNEL *channel
        cdef bytes b_host = to_bytes(host)
        cdef char *_host = b_host
        with nogil:
            channel = c_ssh2.libssh2_channel_direct_tcpip(
                self._session, _host, port)
        if channel is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PyChannel(channel, self)

    def block_directions(self):
        """Get blocked directions for the current session.

        From libssh2 documentation:

        Can be a combination of:

        ``ssh2.session.LIBSSH2_SESSION_BLOCK_INBOUND``: Inbound direction
        blocked.

        ``ssh2.session.LIBSSH2_SESSION_BLOCK_OUTBOUND``: Outbound direction
        blocked.

        Application should wait for data to be available for socket prior to
        calling a libssh2 function again. If ``LIBSSH2_SESSION_BLOCK_INBOUND``
        is set select should contain the session socket in readfds set.

        Correspondingly in case of ``LIBSSH2_SESSION_BLOCK_OUTBOUND`` writefds
        set should contain the socket.

        :rtype: int"""
        cdef int rc
        with nogil:
            rc = c_ssh2.libssh2_session_block_directions(
                self._session)
        return rc

    def forward_listen(self, int port):
        """Create forward listener on port.

        :param port: Port to listen on.
        :type port: int

        :rtype: :py:class:`ssh2.listener.Listener` or None"""
        cdef c_ssh2.LIBSSH2_LISTENER *listener
        with nogil:
            listener = c_ssh2.libssh2_channel_forward_listen(
                self._session, port)
        if listener is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PyListener(listener, self)

    def forward_listen_ex(self, host not None, int port,
                          int bound_port, int queue_maxsize):
        cdef c_ssh2.LIBSSH2_LISTENER *listener
        cdef bytes b_host = to_bytes(host)
        cdef char *_host = b_host
        with nogil:
            listener = c_ssh2.libssh2_channel_forward_listen_ex(
                self._session, _host, port, &bound_port, queue_maxsize)
        if listener is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PyListener(listener, self)

    def sftp_init(self):
        """Initialise SFTP channel.

        :rtype: :py:class:`ssh2.sftp.SFTP`
        """
        cdef c_sftp.LIBSSH2_SFTP *_sftp
        with nogil:
            _sftp = c_sftp.libssh2_sftp_init(self._session)
        if _sftp is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PySFTP(_sftp, self)

    def last_error(self, size_t msg_size=1024):
        """Retrieve last error message from libssh2, if any.
        Returns empty string on no error message.

        :rtype: str
        """
        cdef char *_error_msg
        cdef bytes msg = b''
        cdef int errmsg_len = 0
        with nogil:
            _error_msg = <char *>malloc(sizeof(char) * msg_size)
            c_ssh2.libssh2_session_last_error(
                self._session, &_error_msg, &errmsg_len, 1)
        try:
            if errmsg_len > 0:
                msg = _error_msg[:errmsg_len]
            return msg
        finally:
            free(_error_msg)

    def last_errno(self):
        """Retrieve last error number from libssh2, if any.
        Returns 0 on no last error.

        :rtype: int
        """
        cdef int rc
        with nogil:
            rc = c_ssh2.libssh2_session_last_errno(
                self._session)
        return rc

    def set_last_error(self, int errcode, errmsg not None):
        cdef bytes b_errmsg = to_bytes(errmsg)
        cdef char *_errmsg = b_errmsg
        cdef int rc
        with nogil:
            rc = c_ssh2.libssh2_session_set_last_error(
                self._session, errcode, _errmsg)
        return rc

    def scp_recv(self, path not None):
        """Receive file via SCP.

        Deprecated in favour or recv2 (requires libssh2 >= 1.7).

        :param path: File path to receive.
        :type path: str

        :rtype: tuple(:py:class:`ssh2.channel.Channel`,
          :py:class:`ssh2.statinfo.StatInfo`) or None"""
        cdef bytes b_path = to_bytes(path)
        cdef char *_path = b_path
        cdef StatInfo statinfo = StatInfo()
        cdef c_ssh2.LIBSSH2_CHANNEL *channel
        with nogil:
            channel = c_ssh2.libssh2_scp_recv(
                self._session, _path, statinfo._stat)
        if channel is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PyChannel(channel, self), statinfo

    def scp_recv2(self, path not None):
        """Receive file via SCP.

        Available only on libssh2 >= 1.7.

        :param path: File path to receive.
        :type path: str

        :rtype: tuple(:py:class:`ssh2.channel.Channel`,
          :py:class:`ssh2.fileinfo.FileInfo`) or ``None``"""
        cdef FileInfo fileinfo = FileInfo()
        cdef bytes b_path = to_bytes(path)
        cdef char *_path = b_path
        cdef c_ssh2.LIBSSH2_CHANNEL *channel
        with nogil:
            channel = c_ssh2.libssh2_scp_recv2(
                self._session, _path, fileinfo._stat)
        if channel is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PyChannel(channel, self), fileinfo

    def scp_send(self, path not None, int mode, size_t size):
        """Deprecated in favour of scp_send64. Send file via SCP.

        :param path: Local file path to send.
        :type path: str
        :param mode: File mode.
        :type mode: int
        :param size: size of file
        :type size: int

        :rtype: :py:class:`ssh2.channel.Channel`"""
        cdef bytes b_path = to_bytes(path)
        cdef char *_path = b_path
        cdef c_ssh2.LIBSSH2_CHANNEL *channel
        with nogil:
            channel = c_ssh2.libssh2_scp_send(
                self._session, _path, mode, size)
        if channel is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PyChannel(channel, self)

    def scp_send64(self, path not None, int mode, c_ssh2.libssh2_uint64_t size,
                   time_t mtime, time_t atime):
        """Send file via SCP.

        :param path: Local file path to send.
        :type path: str
        :param mode: File mode.
        :type mode: int
        :param size: size of file
        :type size: int

        :rtype: :py:class:`ssh2.channel.Channel`"""
        cdef bytes b_path = to_bytes(path)
        cdef char *_path = b_path
        cdef c_ssh2.LIBSSH2_CHANNEL *channel
        with nogil:
            channel = c_ssh2.libssh2_scp_send64(
                self._session, _path, mode, size, mtime, atime)
        if channel is NULL:
            return handle_error_codes(c_ssh2.libssh2_session_last_errno(
                self._session))
        return PyChannel(channel, self)

    def publickey_init(self):
        """Initialise public key subsystem for managing remote server
        public keys"""
        cdef c_pkey.LIBSSH2_PUBLICKEY *_pkey
        with nogil:
            _pkey = c_pkey.libssh2_publickey_init(self._session)
        if _pkey is NULL:
            raise PublicKeyInitError
        return PyPublicKeySystem(_pkey, self)

    def hostkey_hash(self, int hash_type):
        """Get computed digest of the remote system's host key.

        :param hash_type: One of ``ssh2.session.LIBSSH2_HOSTKEY_HASH_MD5`` or
          ``ssh2.session.LIBSSH2_HOSTKEY_HASH_SHA1``
        :type hash_type: int

        :rtype: bytes"""
        cdef const char *_hash
        cdef bytes b_hash
        with nogil:
            _hash = c_ssh2.libssh2_hostkey_hash(self._session, hash_type)
        if _hash is NULL:
            return
        b_hash = _hash
        return b_hash

    def hostkey(self):
        """Get server host key for this session.

        Returns key, key_type tuple where key_type is one of
        :py:class:`ssh2.session.LIBSSH2_HOSTKEY_TYPE_RSA`,
        :py:class:`ssh2.session.LIBSSH2_HOSTKEY_TYPE_DSS`, or
        :py:class:`ssh2.session.LIBSSH2_HOSTKEY_TYPE_UNKNOWN`

        :rtype: tuple(bytes, int)"""
        cdef bytes key = b""
        cdef const char *_key
        cdef size_t key_len = 0
        cdef int key_type = 0
        with nogil:
            _key = c_ssh2.libssh2_session_hostkey(
                self._session, &key_len, &key_type)
        if _key is NULL:
            raise SessionHostKeyError(
                "Error retrieving server host key for session")
        key = _key[:key_len]
        return key, key_type

    def knownhost_init(self):
        """Initialise a collection of known hosts for this session.

        :rtype: :py:class:`ssh2.knownhost.KnownHost`"""
        cdef c_ssh2.LIBSSH2_KNOWNHOSTS *known_hosts
        with nogil:
            known_hosts = c_ssh2.libssh2_knownhost_init(
                self._session)
        if known_hosts is NULL:
            raise KnownHostError
        return PyKnownHost(self, known_hosts)

    def keepalive_config(self, bint want_reply, unsigned interval):
        """
        Configure keep alive settings.

        :param want_reply: True/False for reply wanted from server on keep
          alive messages being sent or not.
        :type want_reply: bool
        :param interval: Required keep alive interval. Set to ``0`` to disable
          keepalives.
        :type interval: int"""
        with nogil:
            c_ssh2.libssh2_keepalive_config(self._session, want_reply, interval)

    def keepalive_send(self):
        """Send keepalive.

        Returns seconds remaining before next keep alive should be sent.

        :rtype: int"""
        cdef int seconds = 0
        cdef int c_seconds = 0
        cdef int rc
        with nogil:
            rc = c_ssh2.libssh2_keepalive_send(self._session, &c_seconds)
        handle_error_codes(rc)
        return c_seconds
